/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.maven.aps.modeling.translator;

import java.io.File;
import java.io.IOException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.inteligr8.maven.aps.modeling.util.Index;
import com.inteligr8.maven.aps.modeling.util.ModelUtil;

/**
 * @author brian@inteligr8.com
 */
public class ApsResourceJsonTranslator implements ApsFileTranslator {

	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final Index<Long, String> apsResourceIndex;
	
	/**
	 * This constructor initializes the default translator.
	 * 
	 * @param apsResourceIndex A map of resource IDs to resource names as defined in the APS Service.
	 */
	public ApsResourceJsonTranslator(Index<Long, String> apsResourceIndex) {
		this.apsResourceIndex = apsResourceIndex;
	}
	
	@Override
	public void translateFile(File file, String formName, Long resourceId) throws IOException {
		this.logger.debug("Translating JSON file: {}", file);
		boolean changed = false;
		
		ObjectNode jsonDescriptor = ModelUtil.getInstance().readJson(file, ObjectNode.class);
		JsonNode jsonResourceId = jsonDescriptor.get("resourceId");
		if (jsonResourceId == null || jsonResourceId.isNull()) {
			this.logger.debug("The resource has no ID to translate");
			return;
		}
		
		this.logger.trace("Found ID {} in the '{}' APS resource descriptor", jsonResourceId, formName);
		
		Long apsResourceId = this.apsResourceIndex.getFirstKey(formName);
		if (apsResourceId == null) {
			this.logger.debug("The resource '{}' does not exist in APS; leaving unchanged", formName);
		} else if (!resourceId.equals(apsResourceId)) {
			this.logger.debug("The resource '{}' exists in APS with ID {}; changing descriptor", formName, apsResourceId);
			jsonDescriptor.put("resourceId", apsResourceId);
			changed = true;
		} else {
			this.logger.trace("The resource '{}' ID does not change; leaving unchanged", formName, apsResourceId);
		}
		
		if (changed)
			ModelUtil.getInstance().writeJson(jsonDescriptor, file, false);
	}

}
